(* 
    File:	 Ports.thy
    Theory Name: Ports
    Logic Image: tilco
*)

Ports  =  PPorts +

types
  (* porte di alto livello *)
  'a s_inPort  (* ricezione sincrona *)
  'a s_outPort (* emissione sincrona *)
  'a a_outPort (* emissione asincrona *)
arities
  s_inPort  :: (term) inPort
  s_outPort :: (term) outPort
  a_outPort :: (term) outPort
consts
(* trasmissione sincrona *)
  s_out    :: 'a s_outPort => 'a p_outPort
  s_SendW  :: ['a s_outPort, 'a, tbool,tbool] => tbool
    ("_.!! (_)/ [ (_) ] ;;/ (_)" [50,50,10,10] 9)
  s_noSend :: 'a::outPort => tbool ("_.~!!" [50] 50)

(* ricezione sincrona *)
  s_in        :: 'a s_inPort => 'a p_inPort
  s_ReceiveW  :: ['a s_inPort, tbool, 'a => tbool] => tbool
    ("_.?? [ (_) ] ;;/ (_)" [50,10,10] 9)
  s_noReceive :: 'a::inPort => tbool ("_.~??" [50] 50)
  s_RWait     :: 'a s_inPort => tbool ("_.RWait" [50] 50)
  s_RValue     :: ['a s_inPort, 'a ]=> tbool ("_.RValue _" [50,50] 50)

syntax
  s_Send :: ['a s_outPort, 'a, tbool] => tbool
    ("_.!! (_) ;;/ (_)" [50,50,10] 9)
  s_Receive :: ['a s_inPort, 'a => tbool] => tbool
    ("_.?? ;; (_)" [50,10] 9)
  send :: ['a::outPort, 'b] => tbool
    ("_.(send (_))" [50,50] 50)
  receive :: ['a::inPort, 'b] => tbool
    ("_.(receive (_))" [50,50] 50)
  receiveAck :: 'a::outPort => tbool
    ("_.receiveAck" [50] 50)
  sendAck :: 'a::inPort => tbool
    ("_.sendAck" [50] 50)

translations
  "p.send v"         == "(s_out p).psend v"
  "p.receive v"      == "(s_in p).preceive v"
  "p.receiveAck"     == "(s_out p).preceiveAck"
  "p.sendAck"        == "(s_in p).psendAck"
  
  "p.!! v ;; P"  == "p.!! v [ true ] ;; P"
  "p.?? ;; P" == "p.?? [ true ] ;; P"
defs
  s_RValue_def "(siP.RValue v) == (since' (siP.receive v #& #~ siP.sendAck)
                                            (#~ siP.sendAck))"
  s_RWait_def "(siP.RWait) == (#~ (#? k. siP.RValue k))"
  s_noReceive_def "siP.~?? == (#? W Q. (siP.?? [ W ] ;; Q)) "
 
rules
  s_SendW_rule
    " rule((soP.!! v [ W ] ;; Q) #-->
              (soP.send v) #&
              (until0 (soP.receiveAck #& Q)
                      (#~ soP.receiveAck #& W)) #&
              (soP.receiveAck #|
              (#~ soP.receiveAck #&
                  until (soP.receiveAck)
                        (#~ soP.receiveAck #& #~soP.!!))))"
  s_noSend_rule
    " rule((soP.~!!) #--> (#! k. #~ soP.send k)) "

  s_SendW_unique
    " rule((soP.!! v1 [W1];;P1)#&(soP.!! v2 [W2];;P2) #-->
	      ((const v1) #= (const v2)) #&
	      ((W1#=W2) #& (P1#=P2))@<-inf,inf>)"
  
  s_ReceiveW1_rule
    " rule((siP.?? [ W ] ;; Q) #& siP.RWait #-->
              (until0 (#? k. siP.receive k #&
                             siP.sendAck #& Q k)
                      (#~(#? k. siP.receive k) #& W)) #&
              ((#? k. siP.receive k) #|
              (#~ (#? k. siP.receive k) #& #~siP.sendAck #&
                  until ((#? k. siP.receive k))
                        (#~ (#? k. siP.receive k) #& #~siP.??))))"
  s_ReceiveW2_rule
    " rule((siP.?? [ W ] ;; Q) #& siP.RValue v #-->
              siP.sendAck #& Q v) "
  s_noReceive_rule
    " rule((siP.~??) #--> (#~ siP.sendAck)) "
  
  s_ReceiveW_unique
    " rule((siP.?? [W1];;P1) #& (siP.?? [W2];;P2) #-->
	      ((W1#=W2) #& (#! v. (P1 v) #= (P2 v)))@<-inf,inf>)"  
  s_ReceiveW_noReceive_rule
    " rule((siP.?? [W1];;P1) #& (siP.~??) #--> false)"
end
