(* 
    File:	 Process.thy
    Theory Name: Process
    Logic Image: Integ
*)

Process  =  Ports +

classes
  process    < term
  tComponent < term
types
  ('a, 'b) tVar
       'a  tProp
  ('a, 'b) tSubproc
  ('a, 'b) tSInPort
  ('a, 'b) tSOutPort
arities
  tVar      :: (process,term)    tComponent
  tProp     :: (process)         tComponent
  tSubproc  :: (process,process) tComponent
  tSInPort  :: (process,term)    tComponent
  tSOutPort :: (process,term)    tComponent
consts
  pvar::['a::process, ('a, 'b) tVar]     => 'b tfun      ("_.var _" [50,50] 55)
  ppro::['a::process, 'a tProp]          => tbool        ("_.pro _" [50,50] 55)
  psub::['a::process, ('a, 'b) tSubproc] => 'b           ("_.sub _" [50,60] 55)
  psip::['a::process, ('a, 'b) tSInPort] => 'b s_inPort  ("_.sip _" [51,51] 55)
  psop::['a::process, ('a, 'b) tSOutPort]=> 'b s_outPort ("_.sop _" [51,51] 55)
  
  pstart :: 'a::process => tbool ("_.process'_start" [50] 55)
  system_start :: tbool
syntax
  pdot::['a::process, 'b::tComponent ] => 'c::term       ("_._" [50,60] 55)  
  pcolon::'b::tComponent =>'c::term                      (":_" [60] 55)
  cstart :: tbool (":process'_start")
rules
  uni_start
    "rule (system_start ?! <-inf,inf>)"
  uni_pstart
    "rule (p.process_start ?! <-inf,inf>)"
  pstart_noReceive
    "rule (p.process_start #--> p.sip inP.noReceive @ <-inf,$#0>)"
  pstart_noSend
    "rule (p.process_start #--> p.sop outP.noSend @ <-inf,$#0>)"
end
ML
  fun print_term [] = ()
    | print_term ((Const(c,_))::l) = (print ("Const "^c^"\n"); print_term l)
    | print_term ((Free(c,_))::l) = (print ("Free "^c^"\n"); print_term l)
    | print_term ((Var((c,_),_))::l) = (print ("Var "^c^"\n"); print_term l)
    | print_term ((t1$t2)::l) = (print "$\n";
         print_term([t1]); print_term([t2]); print_term l)
    | print_term (_::l) = (print "???\n"; print_term l);

  val default_process_name = ref "";

  (* inst_of : string -> string
       restituisce la parte del nome di un istanza di un oggetto che indica
       il nome dell'istanza.
        es: inst_of "PROC__inst" = "inst"
  *)   
  fun inst_of s =
    let
      fun strip [] = ""
        | strip [c1] = ""
        | strip (c1 :: (c2 :: l)) =
            if c1="_" andalso c2="_" then
              implode l
            else if c2="_" then
              strip (c2::l)
            else
              strip l;
     in
       strip(explode s)
     end;

  (* proc_of : string -> string
       restituisce la parte del nome di un istanza di un oggetto che indica il
       nome del tipo del processo a cui appartiene.
       es: proc_of "PROC__inst" = "PROC"
  *)  
  fun proc_of s =
    let
      fun proc_in [] = ""
        | proc_in [c] = c
        | proc_in (c1::c2::l) =
          if (c1="_" andalso c2="_") then
            ""
          else
            c1^(proc_in (c2::l))
     in
       proc_in (explode s)
     end;

  val print_process_type= ref false;

  (* strip_proc: string -> string
       se print_process_type e' false
         ritorna il nome dell'istanza senza il nome del processo a cui
         appartiene (elimina dalla stringa s la parte iniziale prima di __)
       se invece e' true
         sostituisce la prima occorrenza di "__" in "::"
       es: strip_proc "PROC__inst" = "inst" o "PROC::inst"
  *)
  fun strip_proc s =
    let
      fun subst [] = ""
        | subst [c1] = c1
        | subst (c1 :: (c2 :: l)) =
            if c1="_" andalso c2="_" then
              ("::" ^ (implode l))
            else if c2="_" then
              c1 ^ (subst (c2::l))
            else
              c1 ^ c2 ^ (subst l);
     in
       if(!print_process_type) then
         subst(explode s)
       else
         inst_of s
     end;

  (* dot_tr': term list -> term
       funzione per la trasformazione in stampa degli operatori ppro, pvar
       psub, psip, psop nell'operatore ".", utilizza strip_proc per togliere
       dai nomi dei componenti i prefissi che indicano il processo a cui
       appartengono.
  *)
  exception ColonFail;
  fun dot_tr' [Const(c1,t1) , Const(c2,t2)] =
        Const("pdot",dummyT)$
        Const(strip_proc(c1),t1) $
        Const(strip_proc(c2),t2)
    | dot_tr' [t , Const(c2,t2)] =
        let
          fun tryColon (Const("_free",t1)$ Free("proc",t2)) =
                Const("pcolon",dummyT)
            | tryColon (Const("_var",t1)$ Var(("proc",0),t2)) =
                Const("pcolon",dummyT)
            | tryColon _ = raise ColonFail
        in
          ((if (proc_of c2) = !default_process_name then
              tryColon t
            else raise ColonFail
           ) handle ColonFail =>
               Const("pdot",dummyT)$ t) $
          (Const(strip_proc(c2),t2))
        end
    | dot_tr' [t , Const(c2,t2) $ x] =
        let
          fun tryColon (Const("_free",t1)$ Free("proc",t2)) =
                Const("pcolon",dummyT)
            | tryColon (Const("_var",t1)$ Var(("proc",0),t2)) =
                Const("pcolon",dummyT)
            | tryColon _ = raise ColonFail
        in
          ((if (proc_of c2) = !default_process_name then
              tryColon t
            else raise ColonFail
           ) handle ColonFail =>
               Const("pdot",dummyT)$ t) $
          (Const(strip_proc(c2),t2) $ x)
        end
    | dot_tr' [t1,t2] =
         (print_term [t1];
         print_term [t2];
         Const("pdot",dummyT)$t1$t2)
    | dot_tr' _ = raise Match;

  fun print_typ (Type(c,tl)) =
      let
        fun pr_typlist [] = print "\n"
         |  pr_typlist (t :: x) = (print_typ t; pr_typlist x)
      in
        (print ("Type: "^c^"\n"); pr_typlist tl)
      end
   | print_typ (TFree(c,s)) = (print ("TFree: "^c^"\n"));
  
  fun pstart_tr' [Const("_free",t1) $ Free("proc",Type(tc,tl))] =
       if tc= !default_process_name then
         Const("cstart",dummyT)
       else raise Match
    | pstart_tr' [Const("_var",t1) $ Var(("proc",n),Type(tc,tl))] =
       if tc= !default_process_name then
         Const("cstart",dummyT)
       else raise Match       
    | pstart_tr' _ = raise Match;
  
  (* viene impostato dot_tr' come funzione di trasformazione per gli operatori
     pvar, ppro, psub, psip, psop
  *)
  val print_translation = [("pvar",dot_tr'),("ppro",dot_tr'),("psub",dot_tr'),
    ("psip",dot_tr'),("psop",dot_tr'),("pstart",pstart_tr')];

  (*************************************************************************)
  
  datatype pComponent = tVar of string
                       | tProp of string
                       | tIn of string
                       | tOut of string
                       | tSub of string*string;
  datatype process = Process of (string*pComponent list);
  datatype process_list = ProcessList of process list;

  (* comp_name : pComponent -> string
       restituisce il nome del componente
  *)
  fun comp_name (tVar n) = n
    | comp_name (tProp n) = n
    | comp_name (tIn n) = n
    | comp_name (tOut n) = n
    | comp_name (tSub (n,_)) = n;

  (* proc_new : string -> process
       ritorna un processo senza componenti con il nome n.
  *)
  fun proc_new n = Process(n,[]);

  infix proc_add;
  (* proc_add : (process * pComponent) -> process
       aggiunge ad un processo un componente, se il nome del componente
       e' gia' presente genera l'eccezione ERROR.
  *)
  fun (Process(n,c)) proc_add v =
    let
      fun add_clist [] v = [v]
        | add_clist (t::l) v =
          if (comp_name t)=(comp_name v) then
            raise ERROR (* componente gia' presente *)
          else
            t::(add_clist l v)
    in
      Process (n,add_clist c v)
    end;

  (* proc_name : process -> string
       restituisce il nome di un processo
  *)
  fun proc_name (Process(n,_)) = n;

  exception ComponentNotFound of (string*string);

  (* proc_find : process -> string -> pComponent
       restituisce il componente di un processo con un certo nome,
       genera l'eccezione ComponentNotFound se il componente non viene trovato
  *)
  fun proc_find (Process(pn,c)) n =
    let
      fun find_clist [] n = raise ComponentNotFound(pn,n)
        | find_clist (r::l) n =
            if (comp_name r)= n then
              r
            else
              (find_clist l n)
    in
      find_clist c n
    end;

  (* procs_new : unit -> process_list
       ritorna una lista di processi vuota
  *)
  fun procs_new () = ProcessList([]);

  (* procs_add : process_list -> process -> process_list
       aggiunge ad una lista di processi un processo, se esiste un processo
       con lo stesso nome viene sostituito.
  *)
  fun procs_add (ProcessList pl) p =
    let
      fun add_plist [] p = [p]
        | add_plist (t::l) p =
          if (proc_name t)=(proc_name p) then
            p::l
          else
            t::(add_plist l p)
    in
      ProcessList (add_plist pl p)
    end;

  exception ProcessNotFound of string;

  (* procs_find : process_list -> string -> process
       ritorna il processo nella lista con un certo nome, genera l'eccezione
       ProcessNotFound se il nome non e' presente.
  *)
  fun procs_find (ProcessList(pl)) n =
    let
      fun find_plist [] n = raise ProcessNotFound(n)
        | find_plist (r::l) n =
            if (proc_name r)= n then
              r
            else
              (find_plist l n)
    in
      find_plist pl n
    end;

  (* default_process_name : ref string
       nome del processo su cui si sta lavorando
     Work_on : string -> unit
       imposta il nome del processo di default
  *)

  fun Work_on pn = default_process_name:=pn;

  (* processes : ref process_list
       e' la lista dei processi del sistema
     Procs_add : process -> unit
       aggiunge un processo a processes, se viene generato ERROR non viene
       fatto niente inoltre viene impostato il processo come default 
     Procs_find : string -> process
       ricerca in processes il processo con un certo nome
  *)
  val processes= ref (procs_new());

  fun Procs_add p = (
    processes := procs_add (!processes) p handle ERROR => ();
    Work_on (proc_name p));
  
  fun Procs_find n = (procs_find (!processes) n);
  
  exception Process_ERROR of string;

  fun print_process_exn e =
    ((case e of
        (Process_ERROR(s)) => print (s^"\n")
      | (ComponentNotFound(pn,c)) =>
          print ("Component "^c^" not found in "^pn^"\n")
      | (ProcessNotFound(pn)) =>
         ((if pn="" then
             print ("Use Work_on to set the default process\n")
           else
             print ("Process "^pn^" not found \n")))
      | _ => raise e);
      raise ERROR);

  (* process_name : term -> string
       restituisce il nome del processo associato a una sequenza di
       sottocomponenti
       es:  p -> default_process_name,
            ((p .sub P__a).sub Q__b) -> tipo processo di Q__b
  *)
  fun process_name (Free(c,_)) =
        !default_process_name
    | process_name (Const("psub",_)$ t $ Const(c,_)) =
      (case (proc_find (Procs_find (proc_of c)) (inst_of c)) of
         tSub(_,n) => n
        | _  => raise Process_ERROR((proc_of c)^"::"^(inst_of c)^
                      " not a subprocess"))
    | process_name _ = raise Process_ERROR("Internal error: 1");

  (* pdot_tr : term list -> term
     funzione che trasforma p.c nella notazione interna:
       se p e' un processo di tipo Q e c e' un componente di Q ad esempio di
       tipo var p.c viene trasformato in p.var Q__c, la traformazione viene
       fatta consultando le tabelle interne relative ai processi.
   *)  
  fun pdot_tr [t,Free(c,_)] =
      let
        val pn=(process_name t)
      in
        (case (proc_find (Procs_find pn) c) of
             tVar(_)  => Const("pvar",dummyT)
           | tProp(_) => Const("ppro",dummyT)
           | tIn(_)   => Const("psip",dummyT)
           | tOut(_)  => Const("psop",dummyT)
           | tSub(_,_)=> Const("psub",dummyT))$
        (t)$
        Const(pn ^ "__" ^ c,dummyT)
        handle e => print_process_exn e
      end
    | pdot_tr [t,Free(c,_) $ x] =
      let
        val pn=(process_name t)
      in
        (case (proc_find (Procs_find pn) c) of
             tVar(_)  => Const("pvar",dummyT)
           | tProp(_) => Const("ppro",dummyT)
           | tIn(_)   => Const("psip",dummyT)
           | tOut(_)  => Const("psop",dummyT)
           | tSub(_,_)=> Const("psub",dummyT))$
        (t)$
        (Const(pn ^ "__" ^ c,dummyT) $ x)
        handle e => print_process_exn e
      end
    | pdot_tr _ = raise Process_ERROR("Internal error: 2");

  (* pcolon_tr : term list -> term
       trasforma :a in proc.S__a dove S e' la stringa impostata con Work_on
   *)
  fun pcolon_tr [Free(c,_)] =
      (((case (proc_find (Procs_find (!default_process_name)) c) of
          tProp(_) => Const("ppro",dummyT)
        | tVar(_) => Const("pvar",dummyT)
        | tIn(_)   => Const("psip",dummyT)
        | tOut(_)  => Const("psop",dummyT)
        | tSub(_,_) => Const("psub",dummyT))$
      Free("proc",dummyT)$
      Const((!default_process_name)^"__"^c,dummyT))
      handle e => print_process_exn e)
    | pcolon_tr [Free(c,_) $ x] =
      (((case (proc_find (Procs_find (!default_process_name)) c) of
          tProp(_) => Const("ppro",dummyT)
        | tVar(_) => Const("pvar",dummyT)
        | tIn(_)   => Const("psip",dummyT)
        | tOut(_)  => Const("psop",dummyT)
        | tSub(_,_) => Const("psub",dummyT))$
      Free("proc",dummyT)$
      (Const((!default_process_name)^"__"^c,dummyT)$ x))
      handle e => print_process_exn e)
    | pcolon_tr _ = raise Process_ERROR("Internal error: 3");

  fun cstart_tr [] = Const("pstart",dummyT) $ Free("proc",dummyT)
    | cstart_tr _  = raise ERROR;
  
  val parse_translation = [("pdot",pdot_tr),("pcolon",pcolon_tr),
                           ("cstart",cstart_tr)];
  
